mod sync_dir;

use std::{env, fs, io, thread, time};
use std::collections::HashMap;
use std::fs::File;
use std::io::{Read, stdin, Write};
use std::path::{Path, PathBuf};
use std::sync::Arc;
use futures_lite::AsyncReadExt;
use tokio_stream::{Stream, StreamExt};
use anyhow;
use std::time::Duration;
use iroh::base::node_addr::AddrInfoOptions;
use iroh::client::Doc;
use iroh::client::docs::{LiveEvent, ShareMode};
use iroh::docs::store::Query;
use iroh::node::Node;
use tokio::sync::Mutex;
use std::sync::mpsc::{channel, TryRecvError};
use iroh::blobs::store::fs::Store;
use iroh::docs::{Author, AuthorId, ContentStatus};
use iroh::net::key;
use iroh::node;
use notify::{DebouncedEvent, RecursiveMode, watcher, Watcher};
use crate::sync_dir::SyncDir;

#[tokio::main]
async fn main() -> anyhow::Result<()> {
    let watch_rate = 100;
    let args: Vec<String> = env::args().skip(1).collect();

    let base_path = env::current_dir().unwrap().join(args[0].parse::<String>().unwrap());
    println!("BASE_PATH: {:?}", base_path);
    let store_path = "./iroh_data/".to_string();

    let sync_dir = if args.len() > 1 {
        SyncDir::new(base_path.clone(), store_path.clone(), Some(args[1].clone())).await?
    }else{
        SyncDir::new(base_path.clone(), store_path.clone(), None).await?
    };
    let sync_dir = Arc::new(Mutex::new(sync_dir));

    let sync_dir_clone = Arc::clone(&sync_dir);
    let sync_dir_lock = sync_dir_clone.lock().await;
    (*sync_dir_lock).load_local_files().await;
    let mut stream = (*sync_dir_lock).doc_subscribe().await?;
    drop(sync_dir_lock);

    let sync_dir_clone_stream = Arc::clone(&sync_dir);
    let stream_task = tokio::spawn(async move {
        let mut keys: Vec<Vec<u8>> = Vec::new();
        while let Some(item) = stream.next().await {
            match item {
                Ok(value) => {
                    println!("{:?}", value);
                    match value {
                        LiveEvent::InsertRemote { content_status, entry, .. } => {
                            if content_status == ContentStatus::Complete {
                                let sync_dir_lock = sync_dir_clone_stream.lock().await;
                                let _ = (*sync_dir_lock).write_new_entry(entry.clone().key().to_vec()).await;
                                drop(sync_dir_lock);
                            }else{
                                keys.push(entry.clone().key().to_vec());
                            }
                        }
                        LiveEvent::ContentReady { .. } => {
                            if !keys.is_empty() {
                                let sync_dir_lock = sync_dir_clone_stream.lock().await;
                                let _ = (*sync_dir_lock).write_new_entry(keys[0].clone()).await;
                                drop(sync_dir_lock);
                                keys.remove(0);
                            } else {
                                println!("Got ContentReady, but the key is empty");
                            }
                        }
                        LiveEvent::PendingContentReady => {
                            for key in &keys {
                                let sync_dir_lock = sync_dir_clone_stream.lock().await;
                                let _ = (*sync_dir_lock).write_new_entry(key.clone()).await;
                                drop(sync_dir_lock);
                            }
                            keys.clear();
                        }
                        _ => {}
                    }
                }
                Err(e) => eprintln!("Error processing stream: {}", e),
            }
        }
    });

    let sync_dir_clone_watcher = Arc::clone(&sync_dir);
    let watcher_task = tokio::spawn(async move {
        println!("Start watching...");
        let (tx, rx) = channel();
        let mut watcher = watcher(tx, Duration::from_secs(10)).unwrap();
        watcher.watch(base_path.clone(), RecursiveMode::Recursive).unwrap();

        loop {
            match rx.try_recv() {
                Ok(event) => {
                    println!("Event: {:?}", event);
                    let sync_dir_lock = sync_dir_clone_watcher.lock().await;
                    match event {
                        DebouncedEvent::Create(path) => {
                            (*sync_dir_lock).get_content_and_load_local_file(path.clone()).await;
                        }
                        DebouncedEvent::Write(path) => {
                            (*sync_dir_lock).get_content_and_load_local_file(path.clone()).await;
                        }
                        DebouncedEvent::Remove(path) => {
                            let _ = (*sync_dir_lock).remove_file(path).await;
                        }
                        DebouncedEvent::Rename(src, dst) => {
                            (*sync_dir_lock).rename_local_file(src, dst).await;
                        }
                        _ => {}
                    }
                    drop(sync_dir_lock);
                }
                Err(e) => match e {
                    TryRecvError::Empty => {
                        tokio::time::sleep(Duration::from_millis(watch_rate.clone())).await;
                    }
                    TryRecvError::Disconnected => {
                        println!("Channel disconnected, stopping watcher.");
                        break;
                    }
                },
            }
        }

        println!("Stop watching");
    });

    
    let mut input = String::new();
    while io::stdin().read_line(&mut input).expect("Failed to read line") > 0 {
        if input.trim() == "q" {
            drop(stream_task);
            drop(watcher_task);
            
            let sync_dir_clone = Arc::clone(&sync_dir);
            let sync_dir_lock = sync_dir_clone.lock().await;
            (*sync_dir_lock).shutdown().await;
            drop(sync_dir_lock);
            break;
        }
    }
    Ok(())
}

