use std::collections::HashMap;
use std::{env, fs};
use std::fs::File;
use std::future::Future;
use std::io::{Read, Write};
use std::path::{Path, PathBuf, StripPrefixError};
use std::ptr::null;
use std::sync::{Arc, MutexGuard};
use std::sync::mpsc::channel;
use std::time::Duration;
use futures_lite::{Stream, StreamExt};
use iroh::base::node_addr::AddrInfoOptions;
use iroh::blobs::store;
use iroh::blobs::store::Store;
use iroh::client::docs::{Client, Entry, LiveEvent, ShareMode};
use iroh::client::{Doc, Iroh};
use iroh::docs::{Author, AuthorId};
use iroh::docs::store::Query;
use iroh::node::Node;
use notify::{DebouncedEvent, RecursiveMode, watcher, Watcher};
use anyhow::{anyhow, Result};
use bytes::Bytes;
use iroh::base::ticket::Ticket;
use serde::{Deserialize, Serialize};
use tokio::fs::remove_file;
use tokio::sync::Mutex;
use tokio::task;

#[derive(Clone, Debug, Serialize, Deserialize)]
pub struct Entity {
    pub content: Vec<u8>,
    pub is_deleted: bool,
}

impl Entity {
    fn new(content: Vec<u8>, is_deleted: bool) -> Self{
        Entity{
            content,
            is_deleted
        }
    }
    
    fn from_bytes(bytes: Bytes) -> anyhow::Result<Self> {
        let todo = serde_json::from_slice(&bytes)?;
        Ok(todo)
    }

    fn as_bytes(&self) -> anyhow::Result<Bytes> {
        let buf = serde_json::to_vec(self)?;
        Ok(buf.into())
    }
}

pub struct SyncDir {
    base_path: PathBuf,
    store_path: String,
    node: Node<store::fs::Store>,
    // node: Node<store::mem::Store>,
    doc: Arc<Mutex<Doc>>,
    author: AuthorId,
}

impl SyncDir{
    pub async fn new(base_path: PathBuf, store_path: String, ticket: Option<String>) -> anyhow::Result<Self>{
        let mut node = iroh::node::Node::persistent(Path::new(&store_path.clone())).await?.spawn().await?;
        // let mut node = iroh::node::Node::memory().spawn().await?;
        let client = node.client();
        
        let doc = match ticket {
            None => {
                let d = Arc::new(Mutex::new(client.docs().create().await?));
                let mut doc_lock = d.lock().await;
                let ticket = (*doc_lock).share(ShareMode::Write, AddrInfoOptions::RelayAndAddresses).await?;
                drop(doc_lock);
                println!("{}", ticket);
                d
            }
            Some(ticket) => {
                Arc::new(Mutex::new(client.docs().import(ticket.parse()?).await.unwrap()))
            }
        };
        
        // let author = client.authors().create().await?;
        let author = client.authors().default().await?;
        
        Ok(SyncDir{
            base_path,
            store_path,
            node,
            doc,
            author
        })
    }
    
    pub async fn load_local_files(&self){
        let directory = self.collect_files_in_dir(self.base_path.clone());
        
        for (key, content) in directory {
            let _ = self.load_local_file(key, content).await;
        }
    }

    pub fn collect_files_in_dir<P: AsRef<Path>>(&self, dir: P) -> HashMap<String, Vec<u8>> {
        let mut file_map = HashMap::new();
        let dir = dir.as_ref();

        if dir.is_dir() {
            for entry in fs::read_dir(dir).expect("Failed to read directory") {
                let entry = entry.expect("Failed to read directory entry");
                let path = entry.path();

                if path.is_dir() {
                    // Recursively collect files in subdirectories
                    file_map.insert(path.clone().to_str().unwrap().to_string(), "".into());
                    file_map.extend(self.collect_files_in_dir(path));
                } else if path.is_file() {
                    // Read the file content and insert it into the map
                    let mut file_content = Vec::new();
                    let mut file = fs::File::open(&path).expect("Failed to open file");
                    file.read_to_end(&mut file_content).expect("Failed to read file");

                    file_map.insert(path.clone().to_str().unwrap().to_string(), file_content);
                }
            }
        }

        file_map
    }

    async fn write_to_doc_if_different(&self, key: String, content: Entity) -> anyhow::Result<()>{
        let entity = self.get_last_from_doc(key.clone()).await;

        match entity {
            Some(entry) => {
                if content.content != entry.content{
                    let _ = self.set_to_doc(key.clone(), content.content.clone(), false).await;
                }
            },
            None => {
                let _ = self.set_to_doc(key.clone(), content.content.clone(), false).await;
            }
        };
        Ok(())
    }

    pub async fn load_local_file(&self, key: String, content: Vec<u8>) -> anyhow::Result<()> {
        let relative_path = self.relative_path(PathBuf::from(key.clone()));

        if content.is_empty() {
            let _ = self.write_to_doc_if_different("d".to_owned() + &*relative_path.display().to_string(), Entity::new(vec![0], false)).await;
        } else {
            let content_clone = content.clone();
            let temp = std::str::from_utf8(&content_clone)?;
            println!("load_local_file: {:?}", temp.clone());
            let _ = self.write_to_doc_if_different("f".to_owned() + &*relative_path.display().to_string(), Entity::new(content.clone(), false)).await;
        }
        
        Ok(())
    }
    
    pub fn remove_local_file(&self, path: PathBuf, is_dir: bool){
        if is_dir{
            let _ = fs::remove_dir_all(path);
        }else{
            let _ = fs::remove_file(path);
        }
    }

    pub async fn write_new_entry(&self, key: Vec<u8>) -> anyhow::Result<()> {
        let entity = self.get_last_from_doc(String::from_utf8(key.clone()).unwrap()).await
            .unwrap_or_else(|| { Entity::new(Vec::new(), false) });

        let cloned_key = String::from_utf8(key)?;
        let mut char_indices = cloned_key.char_indices();
        let second_char_start_index = char_indices.nth(1)
            .map(|(index, _)| index)
            .unwrap_or(cloned_key.len());
        let mut entry_name = self.base_path.clone();
        entry_name = entry_name.join(&cloned_key[second_char_start_index..]);
        
        if entity.is_deleted{
            match cloned_key.chars().next() {
                Some(first_char) if first_char == 'f' => self.remove_local_file(entry_name, false),
                Some(first_char) if first_char == 'd' => self.remove_local_file(entry_name, true),
                Some(_) => {}
                None => {}
            }
            return Ok(())
        }
        
        let content_bytes = Bytes::from(entity.content);        

        println!("write_new_entry: {:?}", entry_name);

        match cloned_key.chars().next() {
            Some(first_char) if first_char == 'f' => {
                if let Some(parent) = entry_name.parent() {
                    fs::create_dir_all(parent)?;
                }
                let mut file = File::create(entry_name)?;
                file.write_all(&*content_bytes)?;
            }
            Some(first_char) if first_char == 'd' => {
                fs::create_dir_all(entry_name)?;
            }
            Some(_) => {}
            None => {}
        }
        
        Ok(())
    }

    pub async fn process_stream(&self, mut stream: impl Stream<Item =anyhow::Result<LiveEvent, anyhow::Error>> + Unpin) -> anyhow::Result<()> {
        let mut keys: Vec<Vec<u8>> = Vec::new();
        while let Some(item) = stream.next().await {
            match item {
                Ok(value) => {
                    println!("{:?}", value);
                    match value {
                        LiveEvent::InsertRemote { entry, .. } => {
                            keys.push(entry.clone().key().to_vec());
                        }
                        LiveEvent::ContentReady { .. } => {
                            if !keys.is_empty() {
                                let _ = self.write_new_entry(keys[0].clone()).await;
                                keys.remove(0);
                            } else {
                                println!("Got ContentReady, but the key is empty");
                            }
                        }
                        LiveEvent::PendingContentReady => {
                            for key in &keys {
                                let _ = self.write_new_entry(key.clone()).await;
                            }
                            keys.clear();
                        }
                        _ => {}
                    }
                }
                Err(e) => eprintln!("Error processing stream: {}", e),
            }
        }

        Ok(())
    }

    pub async fn doc_subscribe(&self) -> Result<impl Stream<Item = Result<LiveEvent>>> {
        let doc_clone = Arc::clone(&self.doc);
        let doc_lock = doc_clone.lock().await;
        let event = doc_lock.subscribe().await;
        drop(doc_lock);
        
        event
    }
    
    pub async fn shutdown(&self) {
        self.node.clone().shutdown().await.expect("Unable to shutdown node");
    }
    
    pub async fn get_by_key(&self, key: String) -> Option<Entity>{
        self.get_last_from_doc(key.clone()).await
    }

    pub async fn get_content_and_load_local_file(&self, path: PathBuf) {
        let mut content: Vec<u8> = vec![];
        if path.is_file() {
            let mut file_content = Vec::new();
            let mut file = fs::File::open(&path).expect("Failed to open file");
            file.read_to_end(&mut file_content).expect("Failed to read file");
            content = file_content;
            let str = String::from_utf8(content.clone()).unwrap();
            println!("local content: {str}");
        }
        self.load_local_file(path.display().to_string(), content).await.expect("Unable to load local file");
    }

    pub async fn remove_if_exists(&self, key: String) -> anyhow::Result<()> {
        let entry = self.get_last_from_doc(key.clone()).await;
        match entry {
            Some(entity) => {
                let _ = self.set_to_doc(key.clone(), entity.content, true).await;
            },
            None => {}
        };


        Ok(())
    }
    
    pub async fn remove_file(&self, path: PathBuf) -> anyhow::Result<()>{
        let relative_path = self.relative_path(path.clone());
        
        let file_path = "f".to_string() + &*relative_path.clone().display().to_string();
        let dir_path = "d".to_string() + &*relative_path.clone().display().to_string();

        // check if file exists
        match self.get_last_from_doc(file_path.clone()).await { 
            Some(entity) if entity.is_deleted != true => {
                let _ = self.remove_if_exists(file_path.clone()).await;
            }
            _ => {} 
        }
        
        // check if directory exists
        match self.get_last_from_doc(dir_path.clone()).await {
            Some(entity) if entity.is_deleted != true => {
                let _ = self.remove_if_exists(dir_path.clone()).await;
            }
            _ => {}
        }
        
        Ok(())
    }
    
    pub fn relative_path(&self, path: PathBuf) -> PathBuf{
        match path.strip_prefix(&self.base_path.clone()) {
            Ok(relative_path) => {PathBuf::from(relative_path)}
            Err(_) => {path}
        }
    }
    
    pub async fn set_to_doc(&self, key: String, content: Vec<u8>, is_deleted: bool) -> anyhow::Result<()>{
        let doc_clone = Arc::clone(&self.doc);
        let mut doc_lock = doc_clone.lock().await;

        let str = String::from_utf8(content.clone()).unwrap();
        println!("SET_TO_DOC: {:?}", str);
        
        let entity = Entity::new(content.clone(), is_deleted.clone());
        (*doc_lock).set_bytes(self.author, key, entity.as_bytes()?).await?;
        drop(doc_lock);
        
        Ok(())
    }
    
    pub async fn get_last_from_doc(&self, key: String) -> Option<Entity> {
        let doc_clone = Arc::clone(&self.doc);
        let mut doc_lock = doc_clone.lock().await;
        let mut entries = (*doc_lock).get_many(Query::single_latest_per_key().key_exact(key.clone())).await.ok()?;
        let entry = entries.next().await;

        
        let content = match entry {
            None => None,
            Some(entry) => {
                match entry { 
                    Ok(entry) => {
                        let content_bytes = entry.content_bytes(&(*doc_lock)).await.ok()?;
                        Some(Entity::from_bytes(content_bytes).unwrap_or(Entity::new(Vec::new(), false)))
                    }
                    Err(_) => {None}
                }
            }
        };
        
        drop(doc_lock);
        
        content
    }
    
    pub async fn rename_local_file(&self, src: PathBuf, dst: PathBuf){
        let _ = self.remove_file(src).await;
        let _ = self.get_content_and_load_local_file(dst).await;
    }
}
